from rest_framework import serializers
from .models import ChunkedUpload
from rest_framework.reverse import reverse


class ChunkedUploadSerializer(serializers.ModelSerializer):
    url = serializers.SerializerMethodField()

    def get_url(self, obj):
        return reverse('chunkedupload-detail',
                       kwargs={'pk': obj.id},
                       request=self.context['request'])

    class Meta:
        model = ChunkedUpload
        fields = '__all__'
        read_only_fields = ('status', 'completed_at')




import time
import os.path
import hashlib
import uuid

from django.db import models, transaction
from django.conf import settings
from django.core.files.uploadedfile import UploadedFile
from django.utils import timezone

from .settings import (
    EXPIRATION_DELTA,
    UPLOAD_PATH,
    STORAGE,
    ABSTRACT_MODEL,
    COMPLETE_EXT,
    INCOMPLETE_EXT,
)

AUTH_USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')


def generate_filename(instance, filename):
    filename = os.path.join(instance.upload_dir, str(instance.id) + INCOMPLETE_EXT)
    return time.strftime(filename)


class ChunkedUpload(models.Model):
    upload_dir = UPLOAD_PATH
    UPLOADING = 1
    COMPLETE = 2
    STATUS_CHOICES = (
        (UPLOADING, 'Incomplete'),
        (COMPLETE, 'Complete'),
    )
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    file = models.FileField(max_length=255,
                            upload_to=generate_filename,
                            storage=STORAGE,
                            null=True)
    filename = models.CharField(max_length=255)
    user = models.ForeignKey(AUTH_USER_MODEL,
                             related_name="%(class)s",
                             editable=False)
    offset = models.BigIntegerField(default=0)
    created_at = models.DateTimeField(auto_now_add=True,
                                      editable=False)
    status = models.PositiveSmallIntegerField(choices=STATUS_CHOICES,
                                              default=UPLOADING)
    completed_at = models.DateTimeField(null=True,
                                        blank=True)

    @property
    def expires_at(self):
        return self.created_at + EXPIRATION_DELTA

    @property
    def expired(self):
        return self.expires_at <= timezone.now()

    @property
    def md5(self, rehash=False):
        if getattr(self, '_md5', None) is None or rehash is True:
            md5 = hashlib.md5()
            self.close_file()
            self.file.open(mode='rb')
            for chunk in self.file.chunks():
                md5.update(chunk)
                self._md5 = md5.hexdigest()
            self.close_file()
        return self._md5

    def delete_file(self):
        if self.file:
            storage, path = self.file.storage, self.file.path
            storage.delete(path)
        self.file = None

    @transaction.atomic
    def delete(self, delete_file=True, *args, **kwargs): 
        super(ChunkedUpload, self).delete(*args, **kwargs)
        if delete_file:
            self.delete_file()
            

    def __unicode__(self):
        return u'<%s - upload_id: %s - bytes: %s - status: %s>' % (
            self.filename, self.id, self.offset, self.status)

    def close_file(self):
        """
        Bug in django 1.4: FieldFile `close` method is not reaching all the
        way to the actual python file.
        Fix: we had to loop all inner files and close them manually.
        """
        file_ = self.file
        while file_ is not None:
            file_.close()
            file_ = getattr(file_, 'file', None)

    def append_chunk(self, chunk, chunk_size=None, save=True):
        self.close_file()
        self.file.open(mode='ab')  # mode = append+binary
        # We can use .read() safely because chunk is already in memory
        self.file.write(chunk.read())
        if chunk_size is not None:
            self.offset += chunk_size
        elif hasattr(chunk, 'size'):
            self.offset += chunk.size
        else:
            self.offset = self.file.size
        self._md5 = None  # Clear cached md5
        if save:
            self.save()
        self.close_file()  # Flush

    def get_uploaded_file(self):
        self.close_file()
        self.file.open(mode='rb')  # mode = read+binary
        return UploadedFile(file=self.file, name=self.filename,
                            size=self.offset)

    @transaction.atomic
    def completed(self, completed_at=timezone.now(), ext=COMPLETE_EXT):
        if ext != INCOMPLETE_EXT:
            original_path = self.file.path
            self.file.name = os.path.splitext(self.file.name)[0] + ext
        self.status = self.COMPLETE
        self.completed_at = completed_at
        self.save()
        if ext != INCOMPLETE_EXT:
            os.rename(
                original_path,
                os.path.splitext(self.file.path)[0] + ext,
            )

    class Meta:
        abstract = ABSTRACT_MODEL




from __future__ import print_function

from collections import Counter
from six import iteritems

import django.apps
from django.core.management.base import BaseCommand
from django.utils import timezone
from django.utils.translation import ugettext as _

from drf_chunked_upload.settings import EXPIRATION_DELTA
from drf_chunked_upload.models import ChunkedUpload


PROMPT_MSG = _(u'Do you want to delete {obj}?')
VALID_RESP = {
    "yes": True,
    "y": True,
    "ye": True,
    "no": False,
    "n": False
}


class Command(BaseCommand):

    # Has to be a ChunkedUpload subclass
    base_model = ChunkedUpload

    help = 'Deletes chunked uploads that have already expired.'

    def add_arguments(self, parser):
        parser.add_argument(
            'models',
            metavar='app.model',
            nargs='*',
            help='Any app.model classes you want to clean up. Default is all ChunkedUpload subclasses within a project.',
        )
        parser.add_argument(
            '-i',
            '--interactive',
            action='store_true',
            dest='interactive',
            default=False,
            help='Prompt for confirmation before each deletion.',
        )
        parser.add_argument(
            '-k',
            '--keep-record',
            action='store_false',
            dest='delete_record',
            default=True,
            help="Don't delete upload records, just uploaded files on disk.",
        )

    def handle(self, *args, **options):
        filter_models = options.get('models', None)
        interactive = options.get('interactive')
        delete_record = options.get('delete_record')

        upload_models = self.get_models(filter_models=filter_models)

        for model in upload_models:
            self.process_model(model, interactive=interactive, delete_record=delete_record)

    def _get_filter_model(self, model):
        model_app, model_name = model.split('.')
        try:
            model_cls = django.apps.apps.get_app_config(model_app).get_model(model_name)
        except LookupError as e:
            print("WARNING: {}", e)
        else:
            if issubclass(model_cls, self.base_model):
                return model_cls
            print("WARNING: Model {} is not a subclass of ChunkedUpload and will be skipped.".format(model))
            return None

    def get_models(self, filter_models=None):
        upload_models = []

        if filter_models:
            # the models were specified and
            # we want to process only them
            for model in filter_models:
                model = self._get_filter_model(model)
                if model:
                    upload_models.append(model)
        else:
            # no models were specified and we want
            # to find all ChunkedUpload classes
            upload_models = \
                [m for m in django.apps.apps.get_models() if issubclass(m, self.base_model)]

        return upload_models

    def process_model(self, model, interactive=False, delete_record=True):
        print('Processing uploads for model {}.{}...'.format(
            model._meta.app_label,
            model.__name__,
        ))

        count = Counter({state[0]: 0 for state in model.STATUS_CHOICES})

        chunked_uploads = model.objects.filter(
            created_at__lt=(timezone.now() - EXPIRATION_DELTA)
        )

        if delete_record == False:
            chunked_uploads = chunked_uploads.exclude(file__isnull=True)

        for chunked_upload in chunked_uploads:
            if interactive and not self.get_confirmation(chunked_upload):
                continue

            count[chunked_upload.status] += 1
            # Deleting objects individually to call delete method explicitly
            if delete_record:
                chunked_upload.delete()
            else:
                chunked_upload.delete_file()
                chunked_upload.save()

        for state, number in iteritems(count):
            print(
                '{} {} upload{}s were deleted.'.format(
                    number,
                    dict(model.STATUS_CHOICES)[state].lower(),
                    (' file' if not delete_record else ''),
                )
            )

    def get_confirmation(self, chunked_upload):
        prompt = PROMPT_MSG.format(obj=chunked_upload) + u' (y/n): '

        while True not in ('y', 'n'):
            answer = VALID_RESP.get(raw_input(prompt).lower(), None)
            if answer is not None:
                return answer



from django.http import HttpResponse

from .settings import ENCODER, CONTENT_TYPE


class Response(HttpResponse):
    """
    """

    def __init__(self, content, status=None, *args, **kwargs):
        super(Response, self).__init__(
            content=ENCODER(content),
            content_type=CONTENT_TYPE,
            status=status,
            *args, **kwargs
        )



from django.contrib import admin

from .models import ChunkedUpload
from .settings import ABSTRACT_MODEL

if not ABSTRACT_MODEL:  # If the model exists

    class ChunkedUploadAdmin(admin.ModelAdmin):
        list_display = ('id', 'filename', 'user', 'status',
                        'created_at')
        search_fields = ('filename',)
        list_filter = ('status',)

    admin.site.register(ChunkedUpload, ChunkedUploadAdmin)



from datetime import timedelta

from django.conf import settings


# How long after creation the upload will expire
DEFAULT_EXPIRATION_DELTA = timedelta(days=1)
EXPIRATION_DELTA = getattr(settings, 'DRF_CHUNKED_UPLOAD_EXPIRATION_DELTA',
                           DEFAULT_EXPIRATION_DELTA)

# Path where uploading files will be stored until completion
DEFAULT_UPLOAD_PATH = 'chunked_uploads/%Y/%m/%d'
UPLOAD_PATH = getattr(settings, 'DRF_CHUNKED_UPLOAD_PATH', DEFAULT_UPLOAD_PATH)

# File extensions for upload files
COMPLETE_EXT = getattr(settings, 'DRF_CHUNKED_UPLOAD_COMPLETE_EXT', '.done')
INCOMPLETE_EXT = getattr(settings, 'DRF_CHUNKED_UPLOAD_INCOMPLETE_EXT', '.part')

# Storage system
STORAGE = getattr(settings, 'DRF_CHUNKED_UPLOAD_STORAGE_CLASS', lambda: None)()

# Boolean that defines if the ChunkedUpload model is abstract or not
ABSTRACT_MODEL = getattr(settings, 'DRF_CHUNKED_UPLOAD_ABSTRACT_MODEL', True)

# Boolean that defines if users beside the creator can access an upload record
USER_RESTRICTED = getattr(settings, "DRF_CHUNKED_UPLOAD_USER_RESTRICED", True)

# Max amount of data (in bytes) that can be uploaded. `None` means no limit
DEFAULT_MAX_BYTES = None
MAX_BYTES = getattr(settings, 'DRF_CHUNKED_UPLOAD_MAX_BYTES', DEFAULT_MAX_BYTES)




"""
Exceptions raised by django-chunked-upload.
"""


class ChunkedUploadError(Exception):
    """
    Exception raised if errors in the request/process.
    """

    def __init__(self, status, **data):
        self.status_code = status
        self.data = data



import re

from rest_framework.generics import GenericAPIView
from rest_framework.response import Response
from rest_framework.mixins import ListModelMixin, RetrieveModelMixin
from rest_framework import status

from django.shortcuts import get_object_or_404
from django.core.files.base import ContentFile

from .settings import MAX_BYTES, USER_RESTRICTED
from .models import ChunkedUpload
from .serializers import ChunkedUploadSerializer
from .exceptions import ChunkedUploadError


class ChunkedUploadBaseView(GenericAPIView):
    """
    Base view for the rest of chunked upload views.
    """

    # Has to be a ChunkedUpload subclass
    model = ChunkedUpload
    serializer_class = ChunkedUploadSerializer

    @property
    def response_serializer_class(self):
        return self.serializer_class

    def get_queryset(self):
        """
        Get (and filter) ChunkedUpload queryset.
        By default, user can only continue uploading his/her own uploads.
        """
        queryset = self.model.objects.all()
        if USER_RESTRICTED:
            if self.request.user.is_authenticated():
                queryset = queryset.filter(user=self.request.user)
        return queryset

    def get_response_data(self, chunked_upload, request):
        """
        Data for the response. Should return a dictionary-like object.
        Called *only* if POST is successful.
        """
        return {}

    def _post(self, request, pk=None, *args, **kwargs):
        raise NotImplementedError

    def _put(self, request, pk=None, *args, **kwargs):
        raise NotImplementedError

    def _get(self, request, pk=None, *args, **kwargs):
        raise NotImplementedError

    def put(self, request, pk=None, *args, **kwargs):
        """
        Handle PUT requests.
        """
        try:
            return self._put(request, pk=pk, *args, **kwargs)
        except ChunkedUploadError as error:
            return Response(error.data, status=error.status_code)

    def post(self, request, pk=None, *args, **kwargs):
        """
        Handle POST requests.
        """
        try:
            return self._post(request, pk=pk, *args, **kwargs)
        except ChunkedUploadError as error:
            return Response(error.data, status=error.status_code)

    def get(self, request, pk=None, *args, **kwargs):
        """
        Handle GET requests.
        """
        try:
            return self._get(request, pk=pk, *args, **kwargs)
        except ChunkedUploadError as error:
            return Response(error.data, status=error.status_code)


class ChunkedUploadView(ListModelMixin, RetrieveModelMixin,
                        ChunkedUploadBaseView):
    """
    Uploads large files in multiple chunks. Also, has the ability to resume
    if the upload is interrupted. PUT without upload ID to create an upload
    and POST to complete the upload. POST with a complete file to upload a
    whole file in one go. Method `on_completion` is a placeholder to
    define what to do when upload is complete.
    """

    # I wouldn't recommend to turn off the md5 check, unless is really
    # impacting your performance. Proceed at your own risk.
    do_md5_check = True

    field_name = 'file'
    content_range_pattern = re.compile(
        r'^bytes (?P<start>\d+)-(?P<end>\d+)/(?P<total>\d+)$'
    )
    max_bytes = MAX_BYTES  # Max amount of data that can be uploaded

    def on_completion(self, chunked_upload, request):
        """
        Placeholder method to define what to do when upload is complete.
        """

    def get_max_bytes(self, request):
        """
        Used to limit the max amount of data that can be uploaded. `None` means
        no limit.
        You can override this to have a custom `max_bytes`, e.g. based on
        logged user.
        """

        return self.max_bytes

    def create_chunked_upload(self, save=False, **attrs):
        """
        Creates new chunked upload instance. Called if no 'id' is
        found in the POST data.
        """
        chunked_upload = self.model(**attrs)
        # file starts empty
        chunked_upload.file.save(name='', content=ContentFile(''), save=save)
        return chunked_upload

    def is_valid_chunked_upload(self, chunked_upload):
        """
        Check if chunked upload has already expired or is already complete.
        """
        if chunked_upload.expired:
            raise ChunkedUploadError(status=status.HTTP_410_GONE,
                                     detail='Upload has expired')
        error_msg = 'Upload has already been marked as "%s"'
        if chunked_upload.status == chunked_upload.COMPLETE:
            raise ChunkedUploadError(status=status.HTTP_400_BAD_REQUEST,
                                     detail=error_msg % 'complete')

    def _put_chunk(self, request, pk=None, whole=False, *args, **kwargs):
        chunk = request.data.get(self.field_name)
        if chunk is None:
            raise ChunkedUploadError(status=status.HTTP_400_BAD_REQUEST,
                                     detail='No chunk file was submitted')

        if pk:
            upload_id = pk
            chunked_upload = get_object_or_404(self.get_queryset(),
                                               pk=upload_id)
            self.is_valid_chunked_upload(chunked_upload)
        else:
            user = request.user if request.user.is_authenticated() else None
            chunked_upload = self.serializer_class(data=request.data)
            if not chunked_upload.is_valid():
                raise ChunkedUploadError(status=status.HTTP_400_BAD_REQUEST,
                                         detail=chunked_upload.errors)
            # chunked_upload is currently a serializer;
            # save returns model instance
            chunked_upload = chunked_upload.save(user=user)

        if whole:
            start = 0
            total = chunk.size
            end = total
        else:
            content_range = request.META.get('HTTP_CONTENT_RANGE', '')
            match = self.content_range_pattern.match(content_range)
            if not match:
                raise ChunkedUploadError(status=status.HTTP_400_BAD_REQUEST,
                                         detail='Error in request headers')

            start = int(match.group('start'))
            end = int(match.group('end'))
            total = int(match.group('total'))

        chunk_size = end - start
        max_bytes = self.get_max_bytes(request)

        if max_bytes is not None and total > max_bytes:
            raise ChunkedUploadError(
                status=status.HTTP_400_BAD_REQUEST,
                detail='Size of file exceeds the limit (%s bytes)' % max_bytes
            )
        if chunked_upload.offset != start:
            raise ChunkedUploadError(status=status.HTTP_400_BAD_REQUEST,
                                     detail='Offsets do not match',
                                     offset=chunked_upload.offset)
        if chunk.size != chunk_size:
            raise ChunkedUploadError(status=status.HTTP_400_BAD_REQUEST,
                                     detail="File size doesn't match headers: file size is {} but {} reported".format(chunk.size, chunk_size))

        chunked_upload.append_chunk(chunk, chunk_size=chunk_size)

        return chunked_upload

    def _put(self, request, pk=None, *args, **kwargs):
        chunked_upload = self._put_chunk(request, pk=pk, *args, **kwargs)
        return Response(
            self.response_serializer_class(chunked_upload,
                                           context={'request': request}).data,
            status=status.HTTP_200_OK
        )

    def md5_check(self, chunked_upload, md5):
        """
        Verify if md5 checksum sent by client matches generated md5.
        """
        if chunked_upload.md5 != md5:
            raise ChunkedUploadError(status=status.HTTP_400_BAD_REQUEST,
                                     detail='md5 checksum does not match')

    def _post(self, request, pk=None, *args, **kwargs):
        chunked_upload = None
        if pk:
            upload_id = pk
        else:
            chunked_upload = self._put_chunk(request, *args,
                                             whole=True, **kwargs)
            upload_id = chunked_upload.id

        md5 = request.data.get('md5')

        error_msg = None
        if self.do_md5_check:
            if not upload_id or not md5:
                error_msg = "Both 'id' and 'md5' are required"
        elif not upload_id:
            error_msg = "'id' is required"
        if error_msg:
            raise ChunkedUploadError(status=status.HTTP_400_BAD_REQUEST,
                                     detail=error_msg)

        if not chunked_upload:
            chunked_upload = get_object_or_404(self.get_queryset(),
                                               pk=upload_id)

        self.is_valid_chunked_upload(chunked_upload)

        if self.do_md5_check:
            self.md5_check(chunked_upload, md5)

        chunked_upload.completed()

        self.on_completion(chunked_upload, request)
        return Response(
            self.response_serializer_class(chunked_upload,
                                           context={'request': request}).data,
            status=status.HTTP_200_OK
        )

    def _get(self, request, pk=None, *args, **kwargs):
        if pk:
            return self.retrieve(request, pk=pk, *args, **kwargs)
        else:
            return self.list(request, *args, **kwargs)




#!/usr/bin/env python

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open('VERSION.txt', 'r') as v:
    version = v.read().strip()

with open('README.rst', 'r') as r:
    readme = r.read()

download_url = (
    'https://github.com/jkeifer/drf-chunked-upload/tarball/%s'
)


setup(
    name='drf-chunked-upload',
    packages=['drf_chunked_upload'],
    version=version,
    description=('Upload large files to Django REST Framework in multiple chunks,' +
                 ' with the ability to resume if the upload is interrupted.'),
    long_description=readme,
    author='Jarrett Keifer',
    author_email='jkeifer0@gmail.com',
    url='https://github.com/jkeifer/drf-chunked-upload',
    download_url=download_url % version,
    install_requires=[],
    license='MIT-Zero'
)
